@@ -1,11 +1,10 @@
-/* Copyright (C) 1998, 1999, 2001 Thorsten Kukuk
+/* Copyright (C) 1998, 1999, 2001, 2014 Thorsten Kukuk
    This file is part of the yp-tools.
    Author: Thorsten Kukuk <kukuk@suse.de>
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   it under the terms of the GNU General Public License version 2 as
+   published by the Free Software Foundation.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -21,15 +20,14 @@
 #include "config.h"
 #endif
 
-#ifdef HAVE_GETOPT_H
 #include <getopt.h>
-#else
-#include "lib/getopt.h"
-#endif
 #include <locale.h>
 #include <libintl.h>
-#include "lib/yp-tools.h"
+#include <stdio.h>
+#include <rpcsvc/ypclnt.h>
+#include <rpcsvc/yp_prot.h>
 #include "lib/nicknames.h"
+#include "lib/yp_all_host.h"
 
 #ifndef _
 #define _(String) gettext (String)
@@ -45,14 +43,14 @@ print_version (void)
 Copyright (C) %s Thorsten Kukuk.\n\
 This is free software; see the source for copying conditions.  There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
-"), "1998");
+"), "2014");
   fprintf (stdout, _("Written by %s.\n"), "Thorsten Kukuk");
 }
 
 static void
 print_usage (FILE *stream)
 {
-  fputs (_("Usage: ypcat [-kt] [-d domain] mapname | -x\n"),
+  fputs (_("Usage: ypcat [-kt] [-d domain] [-h hostname] mapname | -x\n"),
 	 stream);
 }
 
@@ -64,6 +62,8 @@ print_help (void)
 
   fputs (_("  -d domain      Use 'domain' instead of the default domain\n"),
 	 stdout);
+  fputs (_("  -h hostname    Query ypserv on 'hostname' instead the current one\n"),
+	 stdout);
   fputs (_("  -k             Display map keys\n"),
 	 stdout);
   fputs (_("  -t             Inhibits map nickname translation\n"), stdout);
@@ -114,8 +114,9 @@ print_data (int status, char *inkey, int inkeylen, char *inval,
 int
 main (int argc, char **argv)
 {
-  int dflag = 0, mflag = 0, tflag = 0, xflag = 0;
+  int dflag = 0, hflag = 0, mflag = 0, tflag = 0, xflag = 0;
   char *domainname = NULL;
+  char *hostname = NULL;
 
   setlocale (LC_MESSAGES, "");
   setlocale (LC_CTYPE, "");
@@ -134,7 +135,7 @@ main (int argc, char **argv)
         {NULL, 0, NULL, '\0'}
       };
 
-      c = getopt_long (argc, argv, "d:ktx?", long_options, &option_index);
+      c = getopt_long (argc, argv, "d:h:ktx?", long_options, &option_index);
       if (c == (-1))
         break;
       switch (c)
@@ -143,6 +144,10 @@ main (int argc, char **argv)
 	  dflag = 1;
 	  domainname = optarg;
 	  break;
+	case 'h':
+	  hflag = 1;
+	  hostname = optarg;
+	  break;
 	case 'k':
 	  kflag = 1;
 	  break;
@@ -179,7 +184,7 @@ main (int argc, char **argv)
     if (argc == 1)
       mflag = 1;
 
-  if ((xflag && (dflag || mflag || tflag)) || (!xflag && !mflag))
+  if ((xflag && (hflag || dflag || mflag || tflag)) || (!xflag && !mflag))
     {
       print_error ();
       return 1;
@@ -213,7 +218,10 @@ main (int argc, char **argv)
       ypcb.foreach = print_data;
       ypcb.data = NULL;
 
-      res = yp_all (domainname, map, &ypcb);
+      if (hflag)
+	res = yp_all_host (domainname, map, &ypcb, hostname);
+      else
+	res = yp_all (domainname, map, &ypcb);
       switch (res)
 	{
 	case YPERR_SUCCESS:
