@@ -1,4 +1,4 @@
-/* Copyright (C) 1998, 1999, 2000, 2001 Thorsten Kukuk
+/* Copyright (C) 1998, 1999, 2000, 2001, 2002, 2004, 2010, 2016 Thorsten Kukuk
    This file is part of the yp-tools.
    Author: Thorsten Kukuk <kukuk@suse.de>
 
@@ -20,23 +20,15 @@
 #include "config.h"
 #endif
 
-#define _GNU_SOURCE
-
 #define keydat_val dptr
 #define keydat_len dsize
 #define valdat_val dptr
 #define valdat_len dsize
 
-#include <pwd.h>
-#ifdef HAVE_CRYPT_H
 #include <crypt.h>
-#endif
 #include <ctype.h>
-#ifdef HAVE_GETOPT_H
 #include <getopt.h>
-#else
-#include "lib/getopt.h"
-#endif
+#include <errno.h>
 #include <time.h>
 #include <netdb.h>
 #include <limits.h>
@@ -46,18 +38,12 @@
 #include <locale.h>
 #include <libintl.h>
 #include <sys/param.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <fcntl.h>
 #include <rpcsvc/yp_prot.h>
 #include <rpcsvc/ypclnt.h>
-#ifdef HAVE_RPC_CLNT_SOC_H
-#include <rpc/clnt_soc.h>
-#endif
-
-#include "yppasswd.h"
-
-#if defined (__NetBSD__) || (defined(__GLIBC__) && (__GLIBC__ == 2 && __GLIBC_MINOR__ == 0))
-/* <rpc/rpc.h> is missing the prototype */
-int getrpcport(char *host, int prognum, int versnum, int proto);
-#endif
+#include <rpcsvc/yppasswd.h>
 
 #ifndef _
 #define _(String) gettext (String)
@@ -65,6 +51,7 @@ int getrpcport(char *host, int prognum, int versnum, int proto);
 
 /* ok, we're using the crack library */
 #ifdef USE_CRACKLIB
+#include <crack.h>
 #ifndef CRACKLIB_DICTPATH
 #define CRACKLIB_DICTPATH "/usr/lib/cracklib_dict"
 #endif
@@ -160,6 +147,13 @@ getnismaster (char *domainname, const char *progname)
 {
   char *master;
   int port, err;
+#if defined(HAVE_TIRPC) && defined(HAVE_YPBIND3)
+  struct netconfig *nconf;
+  struct netbuf svcaddr;
+  char addrbuf[INET6_ADDRSTRLEN];
+  void *handle;
+  int found;
+#endif
 
   if ((err = yp_master (domainname, "passwd.byname", &master)) != 0)
     {
@@ -167,7 +161,46 @@ getnismaster (char *domainname, const char *progname)
                progname, yperr_string (err));
       return NULL;
     }
+
+#if defined(HAVE_TIRPC) && defined(HAVE_YPBIND3)
+  svcaddr.len = 0;
+  svcaddr.maxlen = sizeof (addrbuf);
+  svcaddr.buf = addrbuf;
+  port = 0;
+  found = 0;
+
+  handle = setnetconfig();
+  while ((nconf = getnetconfig(handle)) != NULL)
+    {
+      if (!strcmp(nconf->nc_proto, "udp"))
+	{
+	  if (rpcb_getaddr(YPPASSWDPROG, YPPASSWDPROC_UPDATE,
+			   nconf, &svcaddr, master))
+	    {
+	      port = taddr2port (nconf, &svcaddr);
+	      endnetconfig (handle);
+	      found=1;
+	      break;
+	    }
+
+	  if (rpc_createerr.cf_stat != RPC_UNKNOWNHOST)
+	    {
+	      clnt_pcreateerror (master);
+	      fprintf (stderr, _("%s: rpcb_getaddr (%s) failed!\n"), progname, master);
+	      return NULL;
+	    }
+	}
+    }
+
+  if (!found)
+    {
+      fprintf (stderr, _("Cannot find suitable transport for protocol 'udp'\n"));
+      return NULL;
+    }
+
+#else
   port = getrpcport (master, YPPASSWDPROG, YPPASSWDPROC_UPDATE, IPPROTO_UDP);
+#endif
   if (port == 0)
     {
       fprintf (stderr,
@@ -248,11 +281,16 @@ ypgetpw (char *master, char *domainname, char *name, int uid)
   CLIENT *clnt;
 
   clnt = clnt_create (master, YPPROG, YPVERS, "udp");
+
+  /* clnt_create can return NULL in some cases */
+  if (clnt == NULL)
+	return NULL;
+
   clnt->cl_auth = authunix_create_default ();
 
   if (name == NULL)
     {
-      if (snprintf (uidbuf, sizeof (uidbuf), "%d", uid) >= sizeof (uidbuf))
+      if (snprintf (uidbuf, sizeof (uidbuf), "%d", uid) >= (int) sizeof (uidbuf))
 	abort ();
 
       keyval = strdup (uidbuf);
@@ -363,11 +401,53 @@ getfield (char *gecos, char *field, int size)
   return sp;
 }
 
-#ifndef USE_CRACKLIB
+#define DES 0
+#define MD5 1
+#define SHA_256 5
+#define SHA_512 6
 
+static int
+get_hash_id (const char *password)
+{
+  int hash_id = DES;
+  if (strncmp(password, "$1$", 3) == 0)
+    hash_id = MD5;
+  else if (strncmp(password, "$5$", 3) == 0)
+    hash_id = SHA_256;
+  else if (strncmp(password, "$6$", 3) == 0)
+    hash_id = SHA_512;
+  return hash_id;
+}
+
+static int
+get_passwd_len (const char *password)
+{
+  static const char *allowed_chars =
+    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
+  int passwdlen = strlen (password);
+  int hash_id = get_hash_id (password);
+
+  /* Some systems (HPU/X) store the password aging info after
+   * the password (with a comma to separate it). To support
+   * this we cut the password after the first invalid char
+   * after the normal 13 ones - in the case of MD5 and DES.
+   * We can't cut at the first invalid char, since MD5
+   * uses $ in the first char. In case of SHA-2 we are looking
+   * for first invalid char after the 38 ones.
+   */
+  if (passwdlen > 13 && (hash_id == DES || hash_id == MD5))
+      passwdlen = 13 + strspn (password + 13, allowed_chars);
+
+  if (passwdlen > 38 && (hash_id == SHA_256 || hash_id == SHA_512))
+      passwdlen = 38 + strspn (password + 38, allowed_chars);
+
+  return passwdlen;
+}
+
+#if ! defined(USE_CRACKLIB) || defined(USE_CRACKLIB_STRICT)
 /* this function will verify the user's password
  * for some silly things.  If we're using cracklib, then
- * this function is unused.
+ * this function is unused unless USE_CRACKLIB_STRICT is set.
  */
 
 static int /* return values: 0 = not ok, 1 = ok */
@@ -375,6 +455,7 @@ verifypassword (struct passwd *pwd, char *pwdstr, uid_t uid)
 {
   char *p, *q;
   int ucase, lcase, other, r;
+  int passwdlen;
 
   if ((strlen (pwdstr) < 6) && uid)
     {
@@ -397,8 +478,10 @@ verifypassword (struct passwd *pwd, char *pwdstr, uid_t uid)
       return 0;
     }
 
-  if (pwd->pw_passwd[0]
-      && !strncmp (pwd->pw_passwd, crypt (pwdstr, pwd->pw_passwd), 13)
+  passwdlen = get_passwd_len (pwd->pw_passwd);
+  /* passwdlen needs to be greater than 1, else crypt could return NULL */
+  if (passwdlen > 1
+      && !strncmp (pwd->pw_passwd, crypt (pwdstr, pwd->pw_passwd), passwdlen)
       && uid)
     {
       fputs (_("You cannot reuse the old password.\n"), stderr);
@@ -432,11 +515,43 @@ verifypassword (struct passwd *pwd, char *pwdstr, uid_t uid)
 
 #endif
 
+#define bin_to_ascii(c) ((c)>=38?((c)-38+'a'):(c)>=12?((c)-12+'A'):(c)+'.')
+
+static void
+create_random_salt (char *salt, int num_chars)
+{
+  int fd;
+  unsigned char c;
+  int i;
+  int res;
+
+  fd = open ("/dev/urandom", O_RDONLY);
+
+  for (i = 0; i < num_chars; i++)
+    {
+      res = 0;
+      if (fd != 0)
+        res = read (fd, &c, 1);
+
+      if (res != 1)
+        c = random ();
+
+      salt[i] = bin_to_ascii (c & 0x3f);
+    }
+
+  salt[num_chars] = 0;
+
+  if (fd != 0)
+    close (fd);
+}
+
 int
 main (int argc, char **argv)
 {
   char *s, *progname, *domainname = NULL, *user = NULL, *master = NULL;
   int f_flag = 0, l_flag = 0, p_flag = 0, error, status;
+  int hash_id = DES;
+  char rounds[11] = "\0"; /* max length is '999999999$' */
   struct yppasswd yppwd;
   struct passwd *pwd;
   CLIENT *clnt;
@@ -447,6 +562,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  srandom (time (NULL));
+
   if ((s = strrchr (argv[0], '/')) != NULL)
     progname = s + 1;
   else
@@ -531,6 +648,21 @@ main (int argc, char **argv)
   else if (argc == 1)
     user = argv[0];
 
+#ifdef YPPASSWD_IS_DEPRECATED
+  if (p_flag)
+    {
+      struct stat st;
+      if (stat (PASSWD_PROG, &st) == 0)
+	{
+	  printf (_("yppasswd is deprecated, use %s instead\n\n"),
+		  PASSWD_PROG);
+	  if (execl (PASSWD_PROG, "passwd", user, NULL) == -1)
+	    fprintf (stderr, _("Calling %s failed: %s\n"), PASSWD_PROG,
+		     strerror (errno));
+	}
+    }
+#endif
+
   if ((error = yp_get_default_domain (&domainname)) != 0)
     {
       fprintf (stderr, _("%s: can't get local yp domain: %s\n"),
@@ -553,7 +685,7 @@ main (int argc, char **argv)
 
   if (user != NULL)
     {
-      if (pwd->pw_uid != uid && uid != 0)
+      if ((uid_t)pwd->pw_uid != uid && uid != 0)
         {
           fprintf (stderr,
 	    _("%s: Only root may change account information for others\n"),
@@ -578,12 +710,12 @@ main (int argc, char **argv)
 
   /* Get old password */
   if ((pwd->pw_passwd != NULL && strlen (pwd->pw_passwd) > 0) ||
-      (pwd->pw_uid != uid))
+      ((uid_t)pwd->pw_uid != uid))
     {
       char prompt[130];
       char *hashpass, *cp;
 
-      if (pwd->pw_uid != uid)
+      if ((uid_t)pwd->pw_uid != uid)
         snprintf (prompt, sizeof (prompt), _("Please enter root password:"));
       else
         snprintf (prompt, sizeof (prompt), _("Please enter %spassword:"),
@@ -593,27 +725,40 @@ main (int argc, char **argv)
       cp = stpcpy (hashpass, "##");
       strcpy (cp, pwd->pw_name);
 
+      hash_id = get_hash_id (pwd->pw_passwd);
+
+      /* Preserve 'rounds=<N>$' (if present) in case of SHA-2 */
+      if (hash_id == SHA_256 || hash_id == SHA_512)
+	{
+	  if (strncmp (pwd->pw_passwd + 3, "rounds=", 7) == 0)
+	    strncpy (rounds, pwd->pw_passwd + 10, strcspn (pwd->pw_passwd + 10, "$") + 1);
+	}
+
       /* We can't check the password with shadow passwords enabled. We
        * leave the checking to yppasswdd */
       if (uid != 0 && strcmp (pwd->pw_passwd, "x") != 0 &&
 	  strcmp (pwd->pw_passwd, hashpass ) != 0)
-        if (strcmp (crypt (s, pwd->pw_passwd), pwd->pw_passwd))
-          {
-            fprintf (stderr, _("Sorry.\n"));
-	    return 1;
-          }
+	{
+	  int passwdlen = get_passwd_len (pwd->pw_passwd);
+	  char *sane_passwd = alloca (passwdlen + 1);
+	  strncpy (sane_passwd, pwd->pw_passwd, passwdlen);
+	  sane_passwd[passwdlen] = 0;
+	  if (strcmp (crypt (s, sane_passwd), sane_passwd))
+	    {
+	      fprintf (stderr, _("Sorry.\n"));
+	      return 1;
+	    }
+	}
       yppwd.oldpass = strdup (s);
     }
 
   if (p_flag)
     {
-#define bin_to_ascii(c) ((c)>=38?((c)-38+'a'):(c)>=12?((c)-12+'A'):(c)+'.')
 #ifdef USE_CRACKLIB
       char *error_msg;
 #endif /* USE_CRACKLIB */
-      char *buf, salt[2], *p = NULL;
+      char *buf, salt[37], *p = NULL;
       int tries = 0;
-      time_t tm;
 
       buf = (char *) malloc (129);
 
@@ -647,10 +792,12 @@ main (int argc, char **argv)
 	      fprintf (stderr, _("Not a valid password: %s.\n"), error_msg);
 	      continue;
 	    }
-#else
+#endif
+
+#if ! defined(USE_CRACKLIB) || defined(USE_CRACKLIB_STRICT)
 	  if (verifypassword (pwd, buf, uid) != 1)
 	    continue;
-#endif /* USE_CRACKLIB */
+#endif
 
 	  p = getpass (_("Please retype new password:"));
 	  if (strcmp (buf, p) == 0)
@@ -662,9 +809,34 @@ main (int argc, char **argv)
 	    }
 	}
 
-      time (&tm);
-      salt[0] = bin_to_ascii (tm & 0x3f);
-      salt[1] = bin_to_ascii ((tm >> 6) & 0x3f);
+      switch (hash_id)
+	{
+	case DES:
+	  create_random_salt (salt, 2);
+	  break;
+
+	case MD5:
+	  /* The user already had a MD5 password, so it's safe to
+	   * use a MD5 password again */
+	  strcpy (salt, "$1$");
+	  create_random_salt (salt + 3, 8);
+	  break;
+
+	case SHA_256:
+	case SHA_512:
+	  /* The user already had a SHA-2 password, so it's safe to
+	   * use a SHA-2 password again */
+	  snprintf (salt, 4, "$%d$", hash_id);
+	  if (strlen (rounds) != 0)
+	    {
+	      strcpy (salt + 3, "rounds=");
+	      strcpy (salt + 3 + 7, rounds);
+	      create_random_salt (salt + 3 + 7 + strlen (rounds), 16);
+	    }
+	  else
+	    create_random_salt (salt + 3, 16);
+	  break;
+	}
 
       yppwd.newpw.pw_passwd = strdup (crypt (buf, salt));
     }
@@ -681,6 +853,7 @@ main (int argc, char **argv)
 	      pwd->pw_name, master);
 
       strncpy (gecos, pwd->pw_gecos, sizeof (gecos));
+      gecos[sizeof (gecos) - 1] = '\0';
       sp = getfield (gecos, oname, sizeof (oname));
       if (newfield (progname, _("Name"), oname, name,
 		    sizeof (name)))
