@@ -1,11 +1,10 @@
-/* Copyright (C) 1998, 1999, 2001 Thorsten Kukuk
+/* Copyright (C) 1998, 1999, 2001, 2014, 2016 Thorsten Kukuk
    This file is part of the yp-tools.
    Author: Thorsten Kukuk <kukuk@suse.de>
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   it under the terms of the GNU General Public License version 2 as
+   published by the Free Software Foundation.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -25,21 +24,23 @@
 #include <netdb.h>
 #include <unistd.h>
 #include <string.h>
-#ifdef HAVE_GETOPT_H
 #include <getopt.h>
-#else
-#include "lib/getopt.h"
-#endif
 #include <locale.h>
 #include <libintl.h>
+#include <arpa/inet.h>
 #include <rpc/rpc.h>
 #include <rpc/pmap_clnt.h>
-#include "lib/yp-tools.h"
+#include <rpcsvc/ypclnt.h>
+#include <rpcsvc/yp_prot.h>
 
 #ifndef _
 #define _(String) gettext (String)
 #endif
 
+#if !defined(HAVE_YPBIND3)
+#define ypbind2_resp ypbind_resp
+#endif
+
 /* Name and version of program.  */
 /* Print the version information.  */
 static void
@@ -50,7 +51,7 @@ print_version (void)
 Copyright (C) %s Thorsten Kukuk.\n\
 This is free software; see the source for copying conditions.  There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
-"), "1998");
+"), "2016");
   /* fprintf (stdout, _("Written by %s.\n"), "Thorsten Kukuk"); */
 }
 
@@ -91,10 +92,8 @@ int
 main (int argc, char **argv)
 {
   static struct timeval RPCTIMEOUT = {25, 0};
-  static struct timeval UDPTIMEOUT = {5, 0};
   char *hostname = NULL, *domainname = NULL, *master = NULL;
-  struct sockaddr_in clnt_saddr;
-  int clnt_sock, result;
+  int result;
   time_t order;
   CLIENT *client;
   bool_t clnt_res;
@@ -167,106 +166,90 @@ main (int argc, char **argv)
 	}
     }
 
-  if (hostname == NULL)
+  if (!hostname)
     {
-      struct ypbind_resp ypbr;
-      struct hostent *hent;
+      int ret;
+#if defined(HAVE_YPBIND3)
+      struct ypbind3_resp yp3_r;
 
-      memset (&clnt_saddr, '\0', sizeof clnt_saddr);
-      clnt_saddr.sin_family = AF_INET;
-      clnt_saddr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
-      clnt_sock = RPC_ANYSOCK;
-      client = clnttcp_create (&clnt_saddr, YPBINDPROG, YPBINDVERS,
-			       &clnt_sock, 0, 0);
-      if (client == NULL)
+      memset (&yp3_r, 0, sizeof (struct ypbind3_resp));
+
+      /* ask local ypbind for NIS server */
+      ret = rpc_call ("localhost", YPBINDPROG, YPBINDVERS, YPBINDPROC_DOMAIN,
+		      (xdrproc_t) xdr_domainname, (caddr_t) &domainname,
+		      (xdrproc_t) xdr_ypbind3_resp, (caddr_t) &yp3_r,
+		      "udp");
+      if (ret == RPC_SUCCESS)
 	{
-	  fputs (_("Couldn't get NIS server"), stderr);
-	  fputs (_(" - ypbind not running ?\n"),
-		 stderr);
-	  return 1;
-	}
-      if (clnt_call (client, YPBINDPROC_DOMAIN,
-		     (xdrproc_t) ytxdr_domainname, (caddr_t) &domainname,
-		     (xdrproc_t) ytxdr_ypbind_resp,
-		     (caddr_t) &ypbr, RPCTIMEOUT) != RPC_SUCCESS)
-	{
-	  clnt_perror (client, _("Couldn't get NIS server"));
-	  clnt_destroy (client);
-	  close (clnt_sock);
-	  return 1;
-	}
-
-      clnt_destroy (client);
-      close (clnt_sock);
-      if (ypbr.ypbind_status != YPBIND_SUCC_VAL)
-	{
-	  fputs (_("Couldn't get NIS server"), stderr);
-	  fputs (": ", stderr);
-	  fputs (ypbinderr_string (ypbr.ypbind_respbody.ypbind_error),
-		 stderr);
-	  fputs ("\n", stderr);
-
-	  return 1;
-	}
-      memset (&clnt_saddr, '\0', sizeof (clnt_saddr));
-      clnt_saddr.sin_family = AF_INET;
-      memcpy (&clnt_saddr.sin_port,
-	      &ypbr.ypbind_respbody.ypbind_bindinfo.ypbind_binding_port,
-	      sizeof (clnt_saddr.sin_port));
-      memcpy (&clnt_saddr.sin_addr.s_addr,
-	      &ypbr.ypbind_respbody.ypbind_bindinfo.ypbind_binding_addr,
-	      sizeof (clnt_saddr.sin_addr.s_addr));
-
-      hent = gethostbyaddr ((char *)&clnt_saddr.sin_addr.s_addr,
-			    sizeof (clnt_saddr.sin_addr.s_addr), AF_INET);
-      hostname = strdup (hent->h_name);
-    }
-  else
-    {
-      struct hostent *hent;
-
-      hent = gethostbyname (hostname);
-      if (!hent)
-	{
-	  switch (h_errno)
+	  if (yp3_r.ypbind_status == YPBIND_SUCC_VAL)
+	    hostname = yp3_r.ypbind3_servername;
+	  else
 	    {
-	    case HOST_NOT_FOUND:
-	      fprintf (stderr, _("Unknown host: %s\n"), hostname);
-	      break;
-	    case TRY_AGAIN:
-	      fprintf (stderr, _("Host name lookup failure\n"));
-	      break;
-	    case NO_DATA:
-	      fprintf (stderr, _("No address associated with name: %s\n"),
-		       hostname);
-	      break;
-	    case NO_RECOVERY:
-	      fprintf (stderr, _("Unknown server error\n"));
-	      break;
-	    default:
-	      fprintf (stderr, _("gethostbyname: Unknown error\n"));
-	      break;
+	      fprintf (stderr, _("Domain not bound\n"));
+	      return 1;
 	    }
+	}
+      else if (ret == RPC_PROGVERSMISMATCH)
+#endif
+	{
+	  /* Looks like ypbind does not support V3 yet, fallback
+	     to V2 */
+	  struct ypbind2_resp yp2_r;
+	  memset (&yp2_r, 0, sizeof (struct ypbind2_resp));
+
+	  /* ask local ypbind for NIS server */
+#if defined (HAVE_YPBIND3)
+	  ret = rpc_call ("localhost", YPBINDPROG, YPBINDVERS_2,
+			  YPBINDPROC_DOMAIN,
+			  (xdrproc_t) xdr_domainname, (caddr_t) &domainname,
+			  (xdrproc_t) xdr_ypbind2_resp, (caddr_t) &yp2_r,
+			  "udp");
+#else
+	  ret = callrpc ("localhost", YPBINDPROG, YPBINDVERS,
+			  YPBINDPROC_DOMAIN,
+			  (xdrproc_t) xdr_domainname, (caddr_t) &domainname,
+			 (xdrproc_t) xdr_ypbind_resp, (caddr_t) &yp2_r);
+#endif
+	  if (ret == RPC_SUCCESS)
+	    {
+	      static char straddr[INET_ADDRSTRLEN];
+	      struct sockaddr_in sa;
+	      sa.sin_family = AF_INET;
+	      sa.sin_addr =
+		yp2_r.ypbind_respbody.ypbind_bindinfo.ypbind_binding_addr;
+
+	      inet_ntop(sa.sin_family, &sa.sin_addr,
+			straddr, sizeof(straddr));
+
+	      hostname = straddr;
+	    }
+	}
+
+      /* Both rpc_call fail */
+      if (ret != RPC_SUCCESS)
+	{
+	  fprintf (stderr, _("Error getting NIS server (%i)\n"), ret);
 	  return 1;
 	}
-      clnt_saddr.sin_family = AF_INET;
-      memcpy (&clnt_saddr.sin_addr.s_addr,hent->h_addr_list[0],
-	      hent->h_length);
-      clnt_saddr.sin_port = htons (pmap_getport (&clnt_saddr,
-						 YPPROG, YPVERS,
-						 IPPROTO_UDP));
     }
 
-  clnt_sock = RPC_ANYSOCK;
-  client = clntudp_create (&clnt_saddr, YPPROG, YPVERS, UDPTIMEOUT,
-			   &clnt_sock);
+  client = clnt_create (hostname, YPPROG, YPVERS, "udp");
   if (client == NULL)
     {
-      fprintf (stderr, _("Can't create connection to %s.\n"), hostname);
+      fprintf (stderr, _("Cannot contact %s, no NIS server running or wrong protocol?\n"),
+	       hostname);
+#if 0
+      /* if we failed, print out all appropriate error messages and exit */
+      fprintf(stderr, "Error calling clnt_create()\n");
+      fprintf(stderr, "PROG: %lu\tVERS: %lu\tNET: %s\n",
+              YPPROG, YPVERS, "udp");
+      fprintf(stderr, "clnt_stat: %d\n", rpc_createerr.cf_stat);
+      fprintf(stderr, "re_errno: %d\n", rpc_createerr.cf_error.re_errno);
+#endif
       return 1;
-    }
+   }
 
-  result = clnt_call (client, YPPROC_DOMAIN, (xdrproc_t) ytxdr_domainname,
+  result = clnt_call (client, YPPROC_DOMAIN, (xdrproc_t) xdr_domainname,
 		      (caddr_t) &domainname,  (xdrproc_t) xdr_bool,
 		      (caddr_t) &clnt_res, RPCTIMEOUT);
   if (result != RPC_SUCCESS)
@@ -280,31 +263,31 @@ main (int argc, char **argv)
   if (clnt_res != TRUE)
     {
       fprintf (stdout, _("Domain %s is not supported by %s.\n"), domainname,
-	       hostname);
+	       hostname ? hostname : "unknown");
       return 1;
     }
 
   req.domain = domainname;
   req.map = argv[0];
   memset (&resp_o, '\0', sizeof (resp_o));
-  res1 = clnt_call (client, YPPROC_ORDER, (xdrproc_t) ytxdr_ypreq_nokey,
-		    (caddr_t) &req, (xdrproc_t) ytxdr_ypresp_order,
+  res1 = clnt_call (client, YPPROC_ORDER, (xdrproc_t) xdr_ypreq_nokey,
+		    (caddr_t) &req, (xdrproc_t) xdr_ypresp_order,
 		    (caddr_t) &resp_o, RPCTIMEOUT);
   if (res1 == 0 && resp_o.status != YP_TRUE)
     res1 = ypprot_err (resp_o.status);
   else
     order = resp_o.ordernum;
-  xdr_free ((xdrproc_t) ytxdr_ypresp_order, (char *) &resp_o);
+  xdr_free ((xdrproc_t) xdr_ypresp_order, (char *) &resp_o);
 
   memset (&resp_m, '\0', sizeof (resp_m));
-  res2 =  clnt_call (client, YPPROC_MASTER, (xdrproc_t) ytxdr_ypreq_nokey,
-		     (caddr_t) &req, (xdrproc_t) ytxdr_ypresp_master,
+  res2 =  clnt_call (client, YPPROC_MASTER, (xdrproc_t) xdr_ypreq_nokey,
+		     (caddr_t) &req, (xdrproc_t) xdr_ypresp_master,
 		     (caddr_t) &resp_m, RPCTIMEOUT);
   if (res2 == 0 && resp_m.status != YP_TRUE)
     res2 = ypprot_err (resp_m.status);
   else
     master = strdup (resp_m.master);
-  xdr_free ((xdrproc_t) ytxdr_ypresp_master, (char *) &resp_m);
+  xdr_free ((xdrproc_t) xdr_ypresp_master, (char *) &resp_m);
 
 
   if (res1 && res2)
